import csv
import io
from collections import OrderedDict

import marshmallow
import marshmallow.fields
import marshmallow.validate

from cl_sii.extras import mm_fields
from cl_sii.libs import mm_utils
from cl_sii.libs import tz_utils


_CSV_ROW_DICT_EXTRA_FIELDS_KEY = None
"""CSV row dict key under which the extra data in the row will be saved."""

_RCV_CSV_EXPECTED_FIELD_NAMES = (
    'Nro',
    'Tipo Doc',
    'Tipo Compra',
    'RUT Proveedor',
    'Razon Social',
    'Folio',
    'Fecha Docto',
    'Fecha Recepcion',
    'Fecha Acuse',
    'Monto Exento',
    'Monto Neto',
    'Monto IVA Recuperable',
    'Monto Iva No Recuperable',
    'Codigo IVA No Rec.',
    'Monto Total',
    'Monto Neto Activo Fijo',
    'IVA Activo Fijo',
    'IVA uso Comun',
    'Impto. Sin Derecho a Credito',
    'IVA No Retenido',
    'Tabacos Puros',
    'Tabacos Cigarrillos',
    'Tabacos Elaborados',
    'NCE o NDE sobre Fact. de Compra',
    'Codigo Otro Impuesto',
    'Valor Otro Impuesto',
    'Tasa Otro Impuesto',
)
_RCV_CSV_DIALECT_KEY = 'sii_rcv'


class _RcvCsvDialect(csv.Dialect):

    """
    CSV dialect of RCV CSV files.

    The properties of this dialect were determined with the help of
    :class:`csv.Sniffer`.

    >>> import gzip
    >>> filename = 'SII-download-RCV-file-http-body-response.csv.gz'
    >>> with gzip.open(filename, 'rt', encoding='utf-8') as f:
    ...     dialect = csv.Sniffer().sniff(f.read(50 * 1024))

    """

    delimiter = ';'
    quotechar = '"'
    escapechar = None
    doublequote = False
    skipinitialspace = False
    lineterminator = '\r\n'
    quoting = csv.QUOTE_MINIMAL


csv.register_dialect(_RCV_CSV_DIALECT_KEY, _RcvCsvDialect)


class RcvCsvRowSchema(marshmallow.Schema):

    EXPECTED_INPUT_FIELDS = tuple(_RCV_CSV_EXPECTED_FIELD_NAMES) + (_CSV_ROW_DICT_EXTRA_FIELDS_KEY, )  # type: ignore  # noqa: E501
    FIELD_FECHA_RECEPCION_DATETIME_TZ = tz_utils.TIMEZONE_CL_SANTIAGO

    class Meta:
        strict = True

    emisor_rut = mm_fields.RutField(
        required=True,
        load_from='RUT Proveedor',
    )
    tipo_dte = marshmallow.fields.Integer(
        required=True,
        load_from='Tipo Doc',
    )
    folio = marshmallow.fields.Integer(
        required=True,
        load_from='Folio',
    )
    fecha_emision_date = mm_utils.CustomMarshmallowDateField(
        format='%d/%m/%Y',  # e.g. '22/10/2018'
        required=True,
        load_from='Fecha Docto',
    )
    fecha_recepcion_datetime = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=True,
        load_from='Fecha Recepcion',
    )
    # note: this field value is set using data passed in the schema context.
    receptor_rut = mm_fields.RutField(
        required=True,
    )
    monto_total = marshmallow.fields.Integer(
        required=True,
        load_from='Monto Total',
    )

    @marshmallow.pre_load
    def preprocess(self, in_data: dict) -> dict:
        # note: required fields checks are run later on automatically thus we may not assume that
        #   values of required fields (`required=True`) exist.

        # Set field value only if it was not in the input data.
        in_data.setdefault('receptor_rut', self.context['receptor_rut'])

        return in_data

    @marshmallow.post_load
    def postprocess(self, data: dict) -> dict:
        # >>> data['fecha_recepcion_datetime'].isoformat()
        # '2018-10-23T01:54:13'
        data['fecha_recepcion_datetime'] = tz_utils.convert_naive_dt_to_tz_aware(
            dt=data['fecha_recepcion_datetime'], tz=self.FIELD_FECHA_RECEPCION_DATETIME_TZ)
        # >>> data['fecha_recepcion_datetime'].isoformat()
        # '2018-10-23T01:54:13-03:00'
        # >>> data['fecha_recepcion_datetime'].astimezone(pytz.UTC).isoformat()
        # '2018-10-23T04:54:13+00:00'

        # note: to express this value in another timezone (but the value does not change), do
        #   `datetime_obj.astimezone(pytz.timezone('some timezone'))`

        return data

    @marshmallow.validates_schema(pass_original=True)
    def validate_schema(self, data: dict, original_data: dict) -> None:
        # Fail validation if there was an unexpected input field.
        unexpected_input_fields = (
            set(original_data)
            - set(self.fields)
            - set(self.EXPECTED_INPUT_FIELDS)
        )
        if unexpected_input_fields:
            raise marshmallow.ValidationError(
                'Unexpected input field', field_names=list(unexpected_input_fields))

    # @marshmallow.validates('field_x')
    # def validate_field_x(self, value):
    #     pass

    ###########################################################################
    # non-marshmallow-related methods
    ###########################################################################

    def deserialize_csv_row(self, row: OrderedDict) -> dict:
        try:
            result = self.load(row)  # type: marshmallow.UnmarshalResult
        except marshmallow.ValidationError as exc:
            exc_msg = "Validation errors during deserialization."
            validation_error_msgs = dict(exc.normalized_messages())
            raise ValueError(exc_msg, validation_error_msgs) from exc

        result_data = result.data  # type: dict
        result_errors = result.errors  # type: dict
        if result_errors:
            raise Exception("Deserialization errors: %s", result_errors)
        return result_data


def create_rcv_csv_reader(
    text_stream: io.TextIOBase,
    expected_fields_strict: bool = True,
) -> csv.DictReader:
    # note: mypy wrongly complains: it does not accept 'fieldnames' to be None but that value
    #   is completely acceptable, and it even is the default!
    #   > error: Argument "fieldnames" to "DictReader" has incompatible type "None"; expected
    #   > "Sequence[str]"
    csv_reader = csv.DictReader(  # type: ignore
        text_stream,
        fieldnames=None,  # the values of the first row will be used as the fieldnames
        restkey=_CSV_ROW_DICT_EXTRA_FIELDS_KEY,
        dialect=_RCV_CSV_DIALECT_KEY,
    )
    if expected_fields_strict and tuple(csv_reader.fieldnames) != _RCV_CSV_EXPECTED_FIELD_NAMES:
        raise Exception(
            "CSV file field names do not match those expected, or their order.",
            csv_reader.fieldnames)

    return csv_reader



"""
SII RCV ("Registro de Compras y Ventas").

.. note::
    The RCV ("Registro de Compras y Ventas") is composed of 2 "registros":
    RC ("Registro de Compras") and RV ("Registro de Ventas").

.. seealso::
    http://www.sii.cl/preguntas_frecuentes/catastro/001_012_6971.htm

"""
import csv
import io
from typing import Callable

from . import parse


def process_rcv_csv_file(
    text_stream: io.TextIOBase,
    rcv_owner_rut: str,
    row_data_handler: Callable,
    max_data_rows: int = None,
) -> int:
    """
    Process a RCV CSV file.

    Processing steps:
    - Create a CSV reader, with auto-detection of header names (first row).
    - Instantiate an schema to parse and deserialize each row.
    - For each data row:
        - Using an appropriate schema, deserialize the raw data.
        - Apply ``row_data_handler`` to the deserialization output.

    :param text_stream: a file-like object, not necessarily a real file
    :param rcv_owner_rut: RCV file owner's RUT
    :param row_data_handler: function be called with parsed row data
    :param max_data_rows: max number of data rows to process (raise exception if exceeded);
        ``None`` means no limit
    :return: number of data rows processed

    """
    # TODO: convert to iterator. That way we do not need the 'row_data_handler' and we can also use
    #   the same function to retrieve the collection of deserialized rows.

    csv_reader = parse.create_rcv_csv_reader(text_stream, expected_fields_strict=True)
    schema = parse.RcvCsvRowSchema(context=dict(receptor_rut=rcv_owner_rut))

    try:
        for row_ix, row_data in enumerate(csv_reader, start=1):
            if max_data_rows is not None and row_ix > max_data_rows:
                # TODO: custom exception
                raise Exception("Exceeded 'max_data_rows' value: {}.".format(max_data_rows))

            try:
                deserialized_row_data = schema.deserialize_csv_row(row_data)
            except Exception as exc:
                exc_msg = "Error deserializing row {} of CSV file: {}".format(row_ix, exc)
                raise Exception(exc_msg) from exc
            try:
                row_data_handler(row_ix, deserialized_row_data)
            except Exception as exc:
                exc_msg = "Error in row_data_handler for row {} of CSV file: {}".format(row_ix, exc)
                raise Exception(exc_msg) from exc

        # The first row in the CSV file is not a data row; it is the headers row.
        rows_processed = csv_reader.line_num - 1
    except csv.Error as exc:
        exc_msg = "CSV error for line {} of CSV file: {}".format(csv_reader.line_num, exc)
        raise Exception(exc_msg) from exc

    return rows_processed



import dataclasses
from dataclasses import field as dc_field
from datetime import date
from typing import Dict, Optional

import cl_sii.contribuyente.constants
import cl_sii.rut.constants
from cl_sii.rut import Rut

from . import constants
from .constants import TipoDteEnum


def validate_dte_folio(value: int) -> None:
    """
    Validate value for DTE field ``folio``.

    :raises ValueError:
    :raises TypeError:

    """
    # note: mypy gets confused and complains about "Unsupported operand types for >/<".
    if (value < constants.DTE_FOLIO_FIELD_MIN_VALUE  # type: ignore
            or value > constants.DTE_FOLIO_FIELD_MAX_VALUE):  # type: ignore
        raise ValueError("Value is out of the valid range for 'folio'.")


def validate_dte_monto_total(value: int) -> None:
    """
    Validate value for DTE field ``monto_total``.

    :raises ValueError:
    :raises TypeError:

    """
    # note: mypy gets confused and complains about "Unsupported operand types for >/<".
    if (value < constants.DTE_MONTO_TOTAL_FIELD_MIN_VALUE  # type: ignore
            or value > constants.DTE_MONTO_TOTAL_FIELD_MAX_VALUE):  # type: ignore
        raise ValueError("Value is out of the valid range for 'monto_total'.")


def validate_contribuyente_razon_social(value: str) -> None:
    """
    Validate value for the "razón social" of a "contribuyente".

    :raises ValueError:
    :raises TypeError:

    """
    if len(value) > len(value.strip()):
        raise ValueError("Value must not have leading or trailing whitespace.")

    if len(value) < 1:
        raise ValueError("Value must not be empty.")

    if len(value) > cl_sii.contribuyente.constants.RAZON_SOCIAL_LONG_MAX_LENGTH:
        raise ValueError("Value exceeds max allowed length.")


@dataclasses.dataclass(frozen=True)
class DteNaturalKey:

    """
    Natural key of a DTE.

    The class instances are immutable.

    This group of fields uniquely identifies a DTE.

    >>> instance = DteNaturalKey(Rut('60910000-1'), TipoDteEnum.FACTURA_ELECTRONICA, 2093465)

    >>> str(instance)
    "DteNaturalKey(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDteEnum.FACTURA_ELECTRONICA: 33>, folio=2093465)"
    >>> str(instance) == repr(instance)
    True
    >>> instance.slug
    '60910000-1--33--2093465'

    """

    emisor_rut: Rut = dc_field()
    """
    RUT of the "emisor" of the DTE.
    """

    tipo_dte: TipoDteEnum = dc_field()
    """
    The kind of DTE.
    """

    folio: int = dc_field()
    """
    The sequential number of a DTE of given kind issued by 'emisor_rut'.
    """

    def __post_init__(self) -> None:
        """
        Run validation automatically after setting the fields values.

        :raises TypeError, ValueError:

        """

        if not isinstance(self.emisor_rut, Rut):
            raise TypeError("Inappropriate type of 'emisor_rut'.")

        if not isinstance(self.tipo_dte, TipoDteEnum):
            raise TypeError("Inappropriate type of 'tipo_dte'.")

        if not isinstance(self.folio, int):
            raise TypeError("Inappropriate type of 'folio'.")

        validate_dte_folio(self.folio)

    def as_dict(self) -> Dict[str, object]:
        return dataclasses.asdict(self)

    @property
    def slug(self) -> str:
        """
        Return an slug representation (that preserves uniquess) of the instance.
        """
        # note: many alternatives were considered and discarded such as:
        #   f'{self.emisor_rut}-{self.tipo_dte}-{self.folio}'
        #   f'{self.emisor_rut}.{self.tipo_dte}.{self.folio}'
        #   f'{self.emisor_rut}/{self.tipo_dte}/{self.folio}'
        #   f'R-{self.emisor_rut}-T-{self.tipo_dte}-F-{self.folio}'
        #   f'rut-{self.emisor_rut}-tipo-{self.tipo_dte}-folio-{self.folio}'

        return f'{self.emisor_rut}--{self.tipo_dte}--{self.folio}'


@dataclasses.dataclass(frozen=True)
class DteDataL0(DteNaturalKey):

    """
    DTE data level 0.

    Its fields are enough to uniquely identify a DTE but nothing more.

    The class instances are immutable.

    >>> instance = DteDataL0(
    ...     Rut('60910000-1'), TipoDteEnum.FACTURA_ELECTRONICA, 2093465, date(2018, 5, 7),
    ...     Rut('60910000-1'), 10403)

    >>> str(instance)
    "DteDataL0(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDteEnum.FACTURA_ELECTRONICA: 33>, " \
    "folio=2093465)"
    >>> str(instance) == repr(instance)
    True
    >>> instance.slug
    '60910000-1--33--2093465'
    >>> instance.natural_key
    "DteNaturalKey(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDteEnum.FACTURA_ELECTRONICA: 33>, folio=2093465)"

    """

    @property
    def natural_key(self) -> DteNaturalKey:
        return DteNaturalKey(emisor_rut=self.emisor_rut, tipo_dte=self.tipo_dte, folio=self.folio)


@dataclasses.dataclass(frozen=True)
class DteDataL1(DteDataL0):

    """
    DTE data level 1.

    It is the minimal set of DTE data fields that are useful.
    For example, SII has an endpoint that confirms that a given DTE exists,
    and the data that it requires can be obtained from this struct.

    The class instances are immutable.

    >>> instance = DteDataL1(
    ...     Rut('60910000-1'), TipoDteEnum.FACTURA_ELECTRONICA, 2093465, date(2018, 5, 7),
    ...     Rut('60910000-1'), 10403)

    >>> str(instance)
    "DteDataL1(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDteEnum.FACTURA_ELECTRONICA: 33>, " \
    "folio=2093465, fecha_emision_date=datetime.date(2018, 5, 7), " \
    "receptor_rut=Rut('60910000-1'), monto_total=10403)"
    >>> str(instance) == repr(instance)
    True

    """

    fecha_emision_date: date = dc_field()
    """
    Field 'fecha_emision' of the DTE.

    .. warning:: It may not match the **real date** on which the DTE was issued
        or received/processed by SII.

    """

    receptor_rut: Rut = dc_field()
    """
    RUT of the "receptor" of the DTE.
    """

    monto_total: int = dc_field()
    """
    Total amount of the DTE.
    """

    def __post_init__(self) -> None:
        """
        Run validation automatically after setting the fields values.

        :raises TypeError, ValueError:

        """
        super().__post_init__()

        if not isinstance(self.fecha_emision_date, date):
            raise TypeError("Inappropriate type of 'fecha_emision_date'.")

        if not isinstance(self.receptor_rut, Rut):
            raise TypeError("Inappropriate type of 'receptor_rut'.")

        if not isinstance(self.monto_total, int):
            raise TypeError("Inappropriate type of 'monto_total'.")

        validate_dte_monto_total(self.monto_total)

    @property
    def natural_key(self) -> DteNaturalKey:
        return DteNaturalKey(emisor_rut=self.emisor_rut, tipo_dte=self.tipo_dte, folio=self.folio)


@dataclasses.dataclass(frozen=True)
class DteDataL2(DteDataL1):

    """
    DTE data level 2.

    About fields
    - ``emisor_razon_social``: redundant but required by the DTE XML schema.
    - ``receptor_razon_social``: redundant but required by the DTE XML schema.
    - ``fecha_vencimiento`` (date): important for some business logic
      but it is not required by the DTE XML schema.

    The class instances are immutable.

    """

    emisor_razon_social: str = dc_field()
    """
    "Razón social" (legal name) of the "emisor" of the DTE.
    """

    receptor_razon_social: str = dc_field()
    """
    "Razón social" (legal name) of the "receptor" of the DTE.
    """

    fecha_vencimiento_date: Optional[date] = dc_field(default=None)
    """
    "Fecha de vencimiento (pago)" of the DTE.
    """

    def __post_init__(self) -> None:
        """
        Run validation automatically after setting the fields values.

        :raises TypeError, ValueError:

        """
        super().__post_init__()

        if not isinstance(self.emisor_razon_social, str):
            raise TypeError("Inappropriate type of 'emisor_razon_social'.")
        validate_contribuyente_razon_social(self.emisor_razon_social)

        if not isinstance(self.receptor_razon_social, str):
            raise TypeError("Inappropriate type of 'receptor_razon_social'.")
        validate_contribuyente_razon_social(self.receptor_razon_social)

        if self.fecha_vencimiento_date is not None:
            if not isinstance(self.fecha_vencimiento_date, date):
                raise TypeError("Inappropriate type of 'fecha_vencimiento_date'.")



"""
DTE-related constants.

Sources: official XML schemas 'SiiTypes_v10.xsd' and 'DTE_v10.xsd'.
https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/

"""
import enum


###############################################################################
# DTE fields / Folio
###############################################################################

# XML element 'DTEDefType/Documento/Encabezado/IdDoc/Folio'
# - description: "Folio del Documento Electronico"
# - XML type: 'FolioType'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/DTE_v10.xsd#L52-L56
# XML type 'FolioType' in official schema 'SiiTypes_v10.xsd'.
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/SiiTypes_v10.xsd#L153-L160

DTE_FOLIO_FIELD_TYPE = int
"""DTE field 'Folio' type."""
DTE_FOLIO_FIELD_MIN_VALUE = 1
"""DTE field 'Folio' min value."""
DTE_FOLIO_FIELD_MAX_VALUE = 10 ** 10
"""DTE field 'Folio' max value."""


###############################################################################
# DTE fields / Monto Total
###############################################################################

# XML element 'DTEDefType/Documento/Encabezado/Totales/MntTotal'
# - description: "Monto Total del DTE"
# - XML type: 'MontoType'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/DTE_v10.xsd#L1160-L1164
# XML type 'MontoType' in official schema 'SiiTypes_v10.xsd'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/SiiTypes_v10.xsd#L563-L570

DTE_MONTO_TOTAL_FIELD_TYPE = int
"""DTE field 'Monto Total' type."""
DTE_MONTO_TOTAL_FIELD_MIN_VALUE = 0
"""DTE field 'Monto Total' min value."""
DTE_MONTO_TOTAL_FIELD_MAX_VALUE = 10 ** 18
"""DTE field 'Monto Total' max value."""


###############################################################################
# DTE fields / Tipo de DTE
###############################################################################

# XML element 'DTEDefType/Documento/Encabezado/IdDoc/TipoDTE'
# - description: "Tipo de DTE"
# - XML type: 'DTEType'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/DTE_v10.xsd#L47-L51

DTE_TIPO_DTE_FIELD_TYPE = int
"""DTE field 'Tipo de DTE' type."""
DTE_TIPO_DTE_FIELD_MIN_VALUE = 1
"""DTE field 'Tipo de DTE' min value."""
# DTE_TIPO_DTE_FIELD_MAX_VALUE = ?
# """DTE field 'Tipo de DTE' max value."""


@enum.unique
class TipoDteEnum(enum.IntEnum):

    """
    Enum of Tipo de DTE.

    Source: XML type ``DTEType`` (enum) in official schema ``SiiTypes_v10.xsd``.
    https://github.com/fyndata/lib-cl-sii-python/blob/8b51350/cl_sii/data/ref/factura_electronica/schema_dte/SiiTypes_v10.xsd#L63-L99

    """

    FACTURA_ELECTRONICA = 33
    """Factura Electrónica."""

    FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA = 34
    """Factura no Afecta o Exenta Electrónica."""
    # aka 'Factura Electrónica de Venta de Bienes y Servicios No afectos o Exento de IVA'

    FACTURA_COMPRA_ELECTRONICA = 46
    """Factura de Compra Electrónica."""
    # Name should have been 'Factura Electrónica de Compra'.

    GUIA_DESPACHO_ELECTRONICA = 52
    """Guía de Despacho Electrónica."""

    NOTA_DEBITO_ELECTRONICA = 56
    """Nota de Débito Electrónica."""

    NOTA_CREDITO_ELECTRONICA = 61
    """Nota de Crédito Electrónica."""



import logging
import os

import defusedxml
import defusedxml.lxml
import lxml.etree
import xml.parsers.expat
import xml.parsers.expat.errors


logger = logging.getLogger(__name__)


###############################################################################
# exceptions
###############################################################################

class BaseXmlParsingError(Exception):

    """
    Base class for all XML parsing errors.
    """


class XmlSyntaxError(BaseXmlParsingError):

    """
    The value to be parsed is syntactically invalid XML.

    It is also possible that some cases of maliciously constructed data are
    reported as syntactically invalid XML e.g. a "billion laughs" attack.

    """


class XmlFeatureForbidden(BaseXmlParsingError):

    """
    The parsed XML contains/uses a feature that is forbidden.

    Usually an XML feature is forbidden for security reasons, to prevent
    some attack vectors.

    .. seealso::
        https://docs.python.org/3/library/xml.html#xml-vulnerabilities

    """


class UnknownXmlParsingError(BaseXmlParsingError):

    """
    An unkwnown XML parsing error or for which there is no handling implementation.

    It is useful because the XML parsing process indirectly uses many
    (standard and 3rd party) libraries, some of them with native
    implementations and/or with a lot of obscure Python magic.

    """


class XmlSchemaDocValidationError(Exception):

    """
    XML document did not be validate against an XML schema.

    """


###############################################################################
# functions
###############################################################################

def parse_untrusted_xml(value: bytes) -> lxml.etree.ElementBase:
    """
    Parse XML-encoded content in value.

    .. note::
        It is ok to use it for parsing untrusted or unauthenticated data.
        See https://docs.python.org/3/library/xml.html#xml-vulnerabilities

    .. warning::
        It is possible that for some cases of maliciously constructed data an
        ``XmlSyntaxError`` will be raised instead of a ``XmlFeatureForbidden``
        exception.

    :raises TypeError:
    :raises XmlSyntaxError: if it is not syntactically valid XML
    :raises XmlFeatureForbidden: if the parsed XML document contains/uses a
        feature that is forbidden
    :raises UnknownXmlParsingError: unkwnown XML parsing error or for which
        there is no handling implementation

    """
    # TODO: limit input max size (it might not be straightforward if value is a generator, which
    #   would be desirable).

    if not isinstance(value, bytes):
        raise TypeError("Value to be parsed as XML must be bytes.")

    # note: with this call, 'defusedxml' will
    # - create a custom parser (instance of 'lxml.etree.XMLParser'), which is what will
    #   fundamentally add safety to the parsing (e.g. using 'defusedxml.lxml.RestrictedElement'
    #   as a custom version of 'lxml.etree.ElementBase'),
    # - call the original 'lxml.etree.fromstring' (binary code),
    # - run 'defusedxml.lxml.check_docinfo'.

    # warning: do NOT change the exception handling order.
    try:

        xml_root_em = defusedxml.lxml.fromstring(
            text=value,
            parser=None,           # default: None (a custom one will be created)
            base_url=None,         # default: None
            forbid_dtd=False,      # default: False (allow Document Type Definition)
            forbid_entities=True,  # default: True (forbid Entity definitions/declarations)
        )  # type: lxml.etree.ElementBase

    except (defusedxml.DTDForbidden,
            defusedxml.EntitiesForbidden,
            defusedxml.ExternalReferenceForbidden) as exc:
        # note: we'd rather use 'defusedxml.DefusedXmlException' but that would catch
        #   'defusedxml.NotSupportedError' as well

        raise XmlFeatureForbidden("XML uses or contains a forbidden feature.") from exc

    except lxml.etree.XMLSyntaxError as exc:
        # note: the MRO of this exception class is:
        # - XMLSyntaxError: "Syntax error while parsing an XML document."
        # - ParseError: "Syntax error while parsing an XML document."
        #   note: do not confuse it with the almost identically named 'lxml.etree.ParserError'
        #   ("Internal lxml parser error"), whose parent class *is not* 'LxmlSyntaxError'.
        # - LxmlSyntaxError: "Base class for all syntax errors."
        # - LxmlError: "Main exception base class for lxml. All other exceptions inherit from
        #   this one.
        # - lxml.etree.Error: "Common base class for all non-exit exceptions."

        # 'exc.msg' is a user-friendly error msg and includes the reference to line and column
        #   e.g. "Detected an entity reference loop, line 1, column 7".
        # Thus we do not need these attributes: (exc.position, exc.lineno, exc.offset)
        exc_msg = "XML syntax error. {}.".format(exc.msg)
        raise XmlSyntaxError(exc_msg) from exc

    except xml.parsers.expat.ExpatError as exc:
        # TODO: if this is reached it means we should improve this exception handler (even if
        #   it is just to raise the same exception with a different message) because
        #   it is a good idea to determine whether the source of the problem really is the
        #   XML-encoded content.

        # https://docs.python.org/3/library/pyexpat.html#expaterror-exceptions
        # https://docs.python.org/3/library/pyexpat.html#xml.parsers.expat.errors.messages
        # e.g.
        #   "unknown encoding"
        #   "mismatched tag"
        #   "parsing aborted"
        #   "out of memory"

        # For sanity crop the XML-encoded content to max 1 KiB (arbitrary value).
        log_msg = "Unexpected XML 'ExpatError' at line {} offset {}: {}. Content: %s".format(
            exc.lineno, exc.offset, xml.parsers.expat.errors.messages[exc.code])
        logger.exception(log_msg, str(value[:1024]))

        exc_msg = "Unexpected error while parsing value as XML. Line {}, offset {}.".format(
            exc.lineno, exc.offset)
        raise UnknownXmlParsingError(exc_msg) from exc

    except lxml.etree.LxmlError as exc:
        # TODO: if this is reached it means we should add another exception handler (even if
        #   it is just to raise the same exception with the same message) because it is a good
        #   idea to determine whether the source of the problem really is the response content.

        # For sanity crop the XML-encoded content to max 1 KiB (arbitrary value).
        log_msg = "Unexpected 'LxmlError' that is not an 'XMLSyntaxError'. Content: %s"
        logger.exception(log_msg, str(value[:1024]))

        exc_msg = "Unexpected error while parsing value as XML."
        raise UnknownXmlParsingError(exc_msg) from exc

    except ValueError as exc:
        # TODO: if this is reached it means we should add another exception handler (even if
        #   it is just to raise the same exception with the same message) because it is a good
        #   idea to determine whether the source of the problem really is the response content.

        # For sanity crop the XML-encoded content to max 1 KiB (arbitrary value).
        log_msg = "Unexpected error while parsing value as XML. Content: %s"
        logger.exception(log_msg, str(value[:1024]))

        exc_msg = "Unexpected error while parsing value as XML."
        raise UnknownXmlParsingError(exc_msg) from exc

    return xml_root_em


def read_xml_schema(filename: str) -> lxml.etree.XMLSchema:
    """
    Instantiate an XML schema object from a file.

    :raises ValueError: if there is no file at ``filename``

    """
    if os.path.exists(filename) and os.path.isfile(filename):
        return lxml.etree.XMLSchema(file=filename)
    raise ValueError("XML schema file not found.", filename)


def validate_xml_doc(xml_schema: lxml.etree.XMLSchema, xml_doc: lxml.etree.ElementBase) -> None:
    """
    Validate ``xml_doc`` against XML schema ``xml_schema``.

    :raises XmlSchemaDocValidationError: if ``xml_doc`` did not be validate
        against ``xml_schema``

    """
    # There are several ways to validate 'xml_doc' according to an 'xml_schema'.
    #   Different calls and what happens if validation passes or fails:
    #   - xml_schema.assert_(xml_doc): nothign / raises 'AssertionError'
    #   - xml_schema.assertValid(xml_doc): nothing / raises 'DocumentInvalid'
    #   - xml_schema.validate(xml_doc): returns True / returns False

    try:
        xml_schema.assertValid(xml_doc)
    except lxml.etree.DocumentInvalid as exc:
        # note: 'exc.error_log' and 'xml_schema.error_log' are the same object
        #   (type 'lxml.etree._ListErrorLog').

        # TODO: advanced error details parsing, without leaking too much information.
        # xml_error_log = exc.error_log  # type: lxml.etree._ListErrorLog
        # last_xml_error = exc.error_log.last_error  # type: lxml.etree._LogEntry
        # last_xml_error_xml_doc_line = last_xml_error.line

        # TODO: does 'xml_schema.error_log' persist? is it necessary to clear it afterwards?
        #   `xml_schema._clear_error_log()`

        # Simplest and safest way to get the error message.
        # Error example:
        #   "Element 'DTE': No matching global declaration available for the validation root., line 2"  # noqa: E501
        validation_error_msg = str(exc)

        raise XmlSchemaDocValidationError(validation_error_msg) from exc



"""
Helpers for parsing DTE data from representations such as XML documents.


Usage:

>>> from cl_sii.dte import parse
>>> from cl_sii.libs import xml_utils

>>> with open('/dir/my_file.xml', mode='rb') as f:
...     xml_doc = xml_utils.parse_untrusted_xml(f.read())

>>> parse.clean_dte_xml(xml_doc)
True
>>> parse.validate_dte_xml(xml_doc)
>>> dte_struct = parse.parse_dte_xml(xml_doc)

"""
import logging
import os
from dataclasses import MISSING, _MISSING_TYPE
from datetime import date
from typing import Optional, Union

import lxml.etree

from cl_sii.libs import xml_utils
from cl_sii.rut import Rut
from . import constants
from . import data_models


logger = logging.getLogger(__name__)


DTE_XMLNS_MAP = {
    'sii-dte': 'http://www.sii.cl/SiiDte',
}
"""
Mapping from XML namespace prefix to full name, for DTE processing.
"""


_DTE_XML_SCHEMA_PATH = os.path.abspath(
    os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        'data/ref/factura_electronica/schema_dte/EnvioDTE_v10.xsd',
    )
)
DTE_XML_SCHEMA_OBJ = xml_utils.read_xml_schema(_DTE_XML_SCHEMA_PATH)
"""
XML schema obj for DTE XML document validation.

It is read from a file at import time to avoid unnecessary reads afterwards.
"""


###############################################################################
# main functions
###############################################################################

def clean_dte_xml(xml_doc: lxml.etree.ElementBase) -> bool:
    """
    Remove some non-compliant (DTE XML schema) data from ``xml_doc``.

    Not all non-compliant data is removed; only some corresponding to popular
    modifications but non-compliant nonetheless.

    The object is modified in-place.

    :returns: whether ``xml_doc`` was modified or not

    """
    modified = False

    xml_etree = xml_doc.getroottree()

    # Remove non-standard but popular element 'DocPersonalizado'.
    xml_em = xml_etree.find('sii-dte:DocPersonalizado', namespaces=DTE_XMLNS_MAP)
    if xml_em is not None:
        modified = True
        xml_doc.remove(xml_em)

    return modified


def validate_dte_xml(xml_doc: lxml.etree.ElementBase) -> None:
    """
    Validate ``xml_doc`` against DTE's XML schema.

    :raises xml_utils.XmlSchemaDocValidationError:

    """
    # TODO: add better and more precise exception handling.
    xml_utils.validate_xml_doc(DTE_XML_SCHEMA_OBJ, xml_doc)


def parse_dte_xml(xml_doc: lxml.etree.ElementBase) -> data_models.DteDataL2:
    """
    Parse and deserialize DTE data from ``xml_doc``.

    """
    # TODO: separate the XML parsing stage from the deserialization stage, which could be
    #   performed by XML-agnostic code (perhaps using Marshmallow or data clacases?).
    #   See :class:`cl_sii.rcv.parse.RcvCsvRowSchema`.

    xml_element_root_tree = xml_doc.getroottree()

    obj_struct = data_models.DteDataL2(
        emisor_rut=_get_emisor_rut(xml_element_root_tree),
        tipo_dte=_get_tipo_dte(xml_element_root_tree),
        folio=_get_folio(xml_element_root_tree),
        fecha_emision_date=_get_fecha_emision(xml_element_root_tree),
        receptor_rut=_get_receptor_rut(xml_element_root_tree),
        monto_total=_get_monto_total(xml_element_root_tree),
        emisor_razon_social=_get_emisor_razon_social(xml_element_root_tree),
        receptor_razon_social=_get_receptor_razon_social(xml_element_root_tree),
        fecha_vencimiento_date=_get_fecha_vencimiento(xml_element_root_tree, default=None),
    )

    return obj_struct


###############################################################################
# helpers
###############################################################################

def _get_tipo_dte(xml_etree: lxml.etree.ElementTree) -> constants.TipoDteEnum:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:IdDoc/sii-dte:TipoDTE'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'TipoDTE' was not found in the XML document.")
    return constants.TipoDteEnum(int(value_str))


def _get_folio(xml_etree: lxml.etree.ElementTree) -> int:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:IdDoc/sii-dte:Folio'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'Folio' was not found in the XML document.")
    return int(value_str)


def _get_fecha_emision(xml_etree: lxml.etree.ElementTree) -> date:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:IdDoc/sii-dte:FchEmis'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'FchEmis' was not found in the XML document.")
    return date.fromisoformat(value_str)


def _get_fecha_vencimiento(
    xml_etree: lxml.etree.ElementTree,
    default: Union[date, None, _MISSING_TYPE] = MISSING,
) -> Optional[date]:

    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:IdDoc/sii-dte:FchVenc'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        if default is None or isinstance(default, date):
            value = default
        elif default is MISSING:
            raise Exception("Element 'FchVenc' was not found in the XML document.")
        else:
            raise TypeError("Invalid type of 'default'.")
    else:
        value = date.fromisoformat(value_str)

    return value


def _get_emisor_rut(xml_etree: lxml.etree.ElementTree) -> Rut:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:Emisor/sii-dte:RUTEmisor'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'RUTEmisor' was not found in the XML document.")
    return Rut(value_str)


def _get_emisor_razon_social(xml_etree: lxml.etree.ElementTree) -> str:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:Emisor/sii-dte:RznSoc'

    value_str: str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'RznSoc' was not found in the XML document.")
    return value_str


def _get_receptor_rut(xml_etree: lxml.etree.ElementTree) -> Rut:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:Receptor/sii-dte:RUTRecep'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'RUTRecep' was not found in the XML document.")
    return Rut(value_str)


def _get_receptor_razon_social(xml_etree: lxml.etree.ElementTree) -> str:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:Receptor/sii-dte:RznSocRecep'

    value_str: str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'RznSocRecep' was not found in the XML document.")
    return value_str


def _get_monto_total(xml_etree: lxml.etree.ElementTree) -> int:
    em_path = 'sii-dte:Documento/sii-dte:Encabezado/sii-dte:Totales/sii-dte:MntTotal'

    value_str = xml_etree.findtext(em_path, namespaces=DTE_XMLNS_MAP)
    if value_str is None:
        raise Exception("Element 'MntTotal' was not found in the XML document.")
    return int(value_str)
